; START.ASM
; PROGETTO: B1601
; Compilatore Avocet X65816

; Codice area F8FE00 - F8FFFF rimappato da banco 0

_START_INC_	.SET	1
_KFR_INC_	.SET	1
_RTC_INC_	.SET	1
_DMA_INC_	.SET	1
_ACIA_INC_	.SET	1
_KBD_INC_	.SET	1
_ADDSER_INC_	.SET	1
_CH375_INC_	.SET	1

	.INCLUDE inc\global.inc
	.INCLUDE inc\dirp00.inc
	.INCLUDE inc\dirp01.inc
	.INCLUDE inc\dirp05.inc

	;.JMPTF8

	.LIST on

	.CODEF8

;----------------------------------------------------------
; ---- CODICE START & INTERRUPT
;----------------------------------------------------------

	.SYSCODE
	
_KbdCnt		.EQU	(DP01ADDR + KbdCnt)
_KbdToggle	.EQU	(DP01ADDR + KbdToggle)
_KbdSt		.EQU	(DP01ADDR + KbdSt)
_PS2Ctl		.EQU	(DP01ADDR + PS2Ctl)
_KbdITail	.EQU	(DP01ADDR + KbdITail)
_KbdIHead	.EQU	(DP01ADDR + KbdIHead)
_KbdShift	.EQU	(DP01ADDR + KbdShift)
_KbdFlag	.EQU	(DP01ADDR + KbdFlag)
	
	LONG_OFF
	
	.GLOBAL _KbdCnt, _KbdToggle, _KbdSt, _PS2Ctl, _KbdITail
	.GLOBAL _KbdIHead, _KbdShift, _KbdFlag
	.GLOBAL TORAM

TORAM:	
; vettori ISR
intsr:
	.DW	int0sr, int1sr, int2sr, int3sr
	.DW	int4sr, int5sr, int6sr, int7sr

int0sr:
	SPISR	0

int1sr:	
	SPISR	1

	
; gestione IRQ2 - VIA 0 (T1, CA2, CB2, CA1)	
int2sr:
	ldy	VIA0+VIAIFR	; flag IFR VIA
	bpl	?50		; no irq da VIA
	tya			; IFR
	asl	a		; N = bit 6 (T1IFRB)
	bpl	?20		; no T1 IRQ
	;tay
	ACC16			; A,M -> 16 bit
	inc	JiffyClk	; incrementa contatore clock 10ms
	bne	?02
	inc	JiffyClk+2
?02:	ACC08			; A,M -> 8 bit
	ldx	#SYSTMRCNT-1	; System Timer (max. 2560ms)
?04:	bit	SysTMF,x	; timer attivo ?
	bpl	?05		; no
	dec	SysTmr,x	; update timer x
	bne	?05
	stz	SysTMF,x	; cancella timer
?05:	dex
	bpl	?04
	lda	VIA0+VIAT1CL	; clear flag T1
?20:	tya			; IFR
	lsr	a		; C  = CA2IFRB
	bcc	?30		; no IRQ CA2 da controller tastiera
	jsr	KbInt		; gestione IRQ tastiera	
	lda	#CA2IFRB	; azzera flag CA2
	sta	VIA0+VIAIFR
	sta	KBCLRIRQ	; azzera linea INT
?30:	tya			; IFR
	and	#CB2IFRB	; CB2-TMF0 abilitato ?
	and	VIA0+VIAIER	;  e Timer motor spirato ?
	beq	?40		; no
	lda	#00010000B
	trb	fdcctl		; flag motor off
	lda	#00001100B
	sta	!FDCDOR		; enable controller, motor off
	lda	#CB2IFRB	; disabilita CB2-TMF0
	sta	VIA0+VIAIER
	sta	VIA0+VIAIFR	; clear flag CB2
?40:
?50:	rts
	
KbInt:
	pea	#DP01ADDR	; DP = pagina 1
	pld
	ldx	KBFR		; primo byte (flag)
	bpl	?20		; bit 7 = 0 -> codice ASCII
	txa
	and	#$B0		; maschera 4 bit alti - ignora release (6)
	cmp	#$80		; codice FLAG ?
	bne	?10		; no
	txa
	and	#$03		; bit 7 e 6 di KbdShift
	lsr	a		; C = bit 6
	ror	a		; C = bit 7
	ror	a
	and	#$C0		; maschera bit 7,6
	sta	KbdShift
	lda	KBFR+1		; flag
	sta	KbdFlag
	rts
?10:	cmp	#$90		; toggle bit ?
	bne	?20		; no
	lda	KBFR+1		; toggle bit
	sta	KbdToggle
	and	#SCROLLLOCKB	; test SCROLL LOCK BIT
	php
	lda	#$80		; SCOLL LOCK bit di ScnScroll
	plp
	beq	?12
	tsb	ScnScroll	; disabilita scroll
	rts
?12:	trb	ScnScroll	; abilita scroll
	rts
?20:	ldy	KbdCnt		; test spazio nel buffer
	cpy	#KBDBUFLEN
	bcs	?90		; no room - scarta tasto
	txa
	bit	#$10		; test 4 bit bassi toggle
	beq	?30		; no toggle
	lda	#$0F		; azzera 4 bit bassi di KbdToggle
	trb	KbdToggle
	txa 
	and	#$0F		; 4 bit bassi di KbdToggle
	tsb	KbdToggle	; imposta 4 bit bassi di KbdToggle
	txa
?30:				; inserisce codice tasto nel buffer
	and	#$C0		; bit 7 -> ASCII(0) oppure CONTROL(1)
				; bit 6 -> MAKE(0) oppure RELEASE(1)
	ldx	KbdITail	; indice coda buffer
	sta	KbdBuf,x	; inserisce flag tasto
	inx
	lda	KBFR+1		; codice tasto
	sta	KbdBuf,x
	inx
	iny			; aggiorna contatore bytes
	iny
	sty	KbdCnt
	cpx	#KBDBUFLEN	; test coda piena
	bcc	?40		; no
	ldx	#0		; coda circolare
?40:	stx	KbdITail	; aggiorna ptr. coda buffer
?90:	rts


; ISR for VIA2/VIA3
; VIA3
; CB2 (neg. edge) : ch375/376 interrupt line
;
; VIA2
; CA1 (negative edge) : interrupt from pic usb host
; CA2 (negative edge) : interrupt from UM245R (usb slave)
; CB1
; CB2 (pos/neg. edge) : acia CTS sensing
; T1 free-run interrupt
int3sr:
	.EXTERN usbdskon, umgetcmd

	lda	!VIA3+VIAIFR
	bpl	?via2
	sta	!VIA3+VIAIFR	; clear interrupt flags
	bit	#CB2IFRB
	beq	?via2		; no usb interrupt
	lda	#CMD_GET_STATUS
	sta	usb0cmd
	nop			; 2uS
	nop
	nop
	nop
	lda	usb0dat
	cmp	#USB_INT_DISCONNECT
	bne	?nxt
	lda	#$7F		; clear all status bits but not bit 7
	trb	usb0st
	bra	?via2		; check via2
?nxt:	cmp	#USB_INT_CONNECT
	bne	?via2
	jsl	usbdskon	
?via2:	pea	#DP05ADDR
	pld
	lda	!VIA2+VIAIFR
	bmi	?go
	rts			; no interrupt from VIA3
?go:	and	!VIA2+VIAIER	; mask off disbled flags
	lsr	a		; <0>: CA2 int. ?
	bcc	?ca1		; no, test bit <1>
	
	; /RXF line from UM245R request service 
	; if UM245 is already plugged to usb host then have data on FIFO
	; if UM245 was previously unplugged then it assert some pulses
	; on /RXF line with dummy data: FF, 1D, 3D, 3D or FF, 3D, 3D
	; last two bytes must be 3D 3D
	;
	tay			; save flags
	ldx	#CA2IFRB	; clear CA2 flag
	stx	.ABS.VIA2+VIAIFR
	bit	usbslv		; UM245 already plugged to usb host?
	bmi	?umrx		; yes so this is incoming data
	
	; sensing a plug in
	;
	ldx	.ABS.UM245R	; read UM245 port and discard data
	bit	usbslv		; first transition?
	bvs	?ca1		; no
	ldx	#$40		; set usb host pending plug-in flag
	stx	usbslv
	ldx	#40		; 600ms timeout
	stx	usbcnt1
	;tya			; restore flags
	bra	?ca1		; go to test CA1 flag

?umrx:	; /RXF line low mean data is available from UM245 FIFO
	; expect 8 bytes
	ldx	#0
?umrxl:	lda	#CA2IFRB	; clear CA2 flag
	sta	!VIA2+VIAIFR
	lda	!UM245R
	sta	usbbuf,x
	inx
	cpx	#8
	bcs	?umrx3
	stz	VIA2+VIAT2CL
	stz	VIA2+VIAT2CH
?umrx2:	lda	#CA2IFRB	; check CA2 flag
	bit	VIA2+VIAIFR
	bne	?umrxl
	lda	#T2IFRB
	bit	VIA2+VIAIFR
	beq	?umrx2
	lda	#$40
	trb	usbum
	bra	?ca1r

;?umrx3:	bit	usbum
;	bmi	?setu		; already connected
	
	; check connession request
;	dex
;?umrx4:	lda	usbbuf,x
;	cmp	.ABS.?usbt,x
;	beq	?umrx5
;	lda	#$55
;	sta	!UM245R
;	bra	?ca1r
;?umrx5: dex
;	bpl	?umrx4
;	lda	#$AA
;	sta	!UM245R
;	lda	#$80
;	tsb	usbum
;	bra	?ca1r
;?setu:	lda	#$40

?umrx3:	lda	#$40
	tsb	usbum		; set bit <6>: fifo data available
?ca1r:	tya			; restore flag
	
?ca1:	lsr	a		; <1>: CA1 int.?
	bcc	?cb2		; no, test bit <3>
	
	; this is pic that request a service (usb host)
	;
	tay			; save flags
	lda	#CA1IFRB	; clear CA1 flag
	sta	!VIA2+VIAIFR
	lda	#$80
	tsb	usbmst		; set usb master flag request
	tya			; restore flag
	
?cb2:	lsr	a		; discard bit <2>
	lsr	a		; <3>: CB2 int.?
	bcc	?cb1		; no, test bit <4>

	; /CTS line of R65C51 changed state (transition)
	;
	tay			; save flags
	lda	#CB2IFRB	; clear CB2 flag
	sta	!VIA2+VIAIFR
	lda	#$01
	tsb	splin3		; transition on /CST line
	ldx	#$40
	lda	#00001000B	; check /CTS line level: PA<3>
	bit	!VIA2+VIAPRANH
	beq	?ctsl		; was a negative transition: /CTS is low

	; positive edge transition on /CTS line: /CTS line is high
	;
	txa
	trb	!VIA2+VIAPCR	; next time sense a negative edge on /CTS line
	lda	#$80
	tsb	splin3		; now /CTS line is high
	bra	?hsk		; check handshake mode for R65C51 (CF=1 here)
	
?ctsl:	; negative edge transition on /CTS line: /CTS line is low
	;
	txa
	tsb	!VIA2+VIAPCR	; next time sense a positive edge on /CTS line
	lda	#$80
	trb	splin3		; now /CTS line is low
	clc			; CF=0: resume TX

?hsk:	; if 65C51 is in hardware handshake mode then must handle
	; local pause flag and TX interrupt state
	;
	bit	spmode3		; test for active handshake
	bpl	?cnt		; no handshake, so ignore CTS trans.
	bvc	?cnt		; softw. handshake, so ignore CTS trans.
	txa
	bcc	?clp		; /CTS low clear local pause and resume TX
	tsb	sppause3	; /CTS high set local pause and stop TX
	bra	?cnt
?clp:	trb	sppause3	
?cnt:	tya			; restore flag

?cb1:	lsr	a		; <4>: CB1 int.?
	bcc	?t1		; no, test bit <6>
	ldx	.ABS.VIA2+VIAPRB
	
?t1:	lsr	a		; discadrd bit <5>
	lsr	a		; <6>: T1 int.?
	bcc	?end		; no

	; T1 used for check UM245 /TXE line: if this line is high
	; for long time this mean that usb host was unplugged,
	; and used for check /DSR line when ACIA interrupt is disabled
	; T1 timeout = 15ms
	lda	!VIA2+VIAT1CL	; clear T1 flag
	bit	usbslv		; plug-in pending?
	bvc	?t2		; no
	dec	usbcnt1		; update counter
	bne	?t2		; timeout not expired
	lda	#$80
	sta	usbslv		; assume UM245 plugged-in
?t2:	lda	!ACIACMD
	lsr	a		; check bit <0>
	bcs	?usb		; ACIA interrupt is enabled
	lda	!ACIASR		; is safe to read ACIA status now
	and	#01000000B	; mask off all bits but /DSR
	beq	?dsr		; clear /DSR flag
	tsb	splin3		; set /DSR flag
	bra	?usb
?dsr:	lda	#$40
	trb	splin3	
?usb:	bit	usbslv
	bpl	?txel		; usb unplugged: nothing to do
	bit	!VIA2+VIAPRANH	; check /TXE
	bvc	?txel		; /TXE is low: clear timeout
	lda	usbtim		; timeout started?
	bne	?dec		; yes, update counter
	lda	#44		; set 660ms (44 x 15) timeout
	sta	usbtim
	bra	?end
?dec:	dec	usbtim		; update counter
	bne	?end		; no timeout
	stz	usbslv		; assume cable unplugged
?txel:	stz	usbtim		; clear any pending timeout	
	
;?umrx:	ldx	usbcnt
;?lp1:	lda	!VIA2+VIAPRA	; check /RXF line
;	cmp	!VIA2+VIAPRA
;	bne	?lp1
;	tay
;	asl	a
;	bcs	?xx
;	lda	!UM245R		; read um245 port
;	sta	!$8000,x
;	tya
;	sta	!$8100,x
;	inx
;?xx:	stx	usbcnt

	bit	usbum
	bvc	?end
	jsl	umgetcmd
?end:
	rts

?usbt:	.DB	$87, $E9, $5D, $93, $B7, $57, $7D, $3B

; serial test board interrupt from 16C550 & 65C51
int4sr:
	pea	#DP05ADDR
	pld
	lda	!ACIASR
	bmi	?acia
	jmp	uartisr		; 16C550 interrupt?
?acia:	bit	#ACIARDRF	; rx data?
	beq	?lin		; no, check lines
	tay			; Y = status
	ldx	.ABS.ACIADR	; X = received data
	and	#00000111B	; any rx error pending?
	beq	?rx		; no
	asl	a		; translate overrun (bit 3) to bit 0
	bit	#00001000B	; test overrun
	beq	$+4		; skip next instruction if not overrun
	ora	#$01		; overrun bit (bit<0>)
	ora	#$80		; set x error bit
	bra	?sst		; set status reg. & discard received data
?rx:	txa			; A = received data
	INDEX16
	ldx	ibufcnt3	; count of bytes stored in input queue
	bit	spmode3		; test for active handshake
	bpl	?chk		; no handshake: goto to check input queue
	bvs	?tst		; hardware handshake (check queue overflow)
	cmp	#SPXON		; received an XON control byte?
	bne	?xoff		; no, check if received an XOFF
	lda	#$40		; clear local pause flag
	trb	sppause3	; bit<6> = 0 -> local pause off (resume tx)
	bra	?cnt		; discard received data
?xoff:	cmp	#SPXOFF		; received an XOFF control byte?
	bne	?tst		; no, so check queue overflow
	lda	#$40		; set local pause flag
	tsb	sppause3	; bit<6> = 1 -> local pause on (stop tx)
	bra	?cnt		; discard received data
?tst:	cpx	icntmax3	; check input buff. for remote pause condition
	bcc	?sto		; below guard limit: store data
	bit	sppause3	; remote pause is already on ?
	bmi	?chk		; yes, so check input buffer
	xba			; B = received data
	bit	spmode3		; test handshake type
	bvs	?rtsh		; bit 6=1 -> hardware handshake so set RTS=1
	lda	#SPXOFF		; send an XOFF to remote terminal
	sta	spout3		; XOFF sending deffered until TDRE is set
	bra	?xba		; check input buffer		
?rtsh:	lda	#$04		; hardware handshake...
	tsb	!VIA2+VIAPRANH	; ...set RTS=1
	lda	#$80		; set remote pause
	tsb	sppause3	; bit 7=1 -> remote pause on
?xba:	xba			; A = received data
?chk:	cpx	#SIBUFSIZ3	; left room in input buffer?
	bcc	?sto		; yes, store received byte
	lda	#$C0		; set bit 7: rx error, bit 6: rx overflow
?sst:	sta	spstat3		; set status register
	bra	?cnt		; discard received data
?sto:	inx			; now store received data
	stx	ibufcnt3	; update bytes count
	ldx	ibuftail3	; pointer to rx tail queue
	sta	>SIBUFADDR3,x
	inx			; update tail pointer
	CPU16
	.MLIST
	txa
	and	#(SIBUFSIZ3-1)	; circular queue
	sta	ibuftail3
?cnt:	CPU08			; continue...	
	tya			; A = acia status
?lin:	tay	
	and	#01000000B	; mask off all bits but /DSR
	beq	?dsr		; clear /DSR flag
	tsb	splin3		; set /DSR flag
	bra	?cnt2
?dsr:	lda	#$40
	trb	splin3	
?cnt2:	tya	
	bit	#ACIATDRE	; TX reg. empty?
	beq	?end		; no
	ldx	spout3		; pending an XON/XOFF sending?
	beq	?cklp		; no... check local pause flag
	
	; if send an XON then we will re-enable tx interrupt
	; else we will disable tx interrupt
	;
	lda	#00001100B	; disable tx for now (is safe)
	trb	!ACIACMD
	cpx	#SPXON
	bne	?di
	lda	#00000100B	; re-enable tx interrupt
	bra	?cmd
?di:	lda	#00001000B	; disable tx interrupt
?cmd:	tsb	!ACIACMD
	stx	.ABS.ACIADR	; send XON/XOFF
	stz	spout3		; clear XON/XOFF flag	
	lda	#$80		; remote pause flag
	cpx	#SPXOFF
	beq	?srp		; sent an XOFF: set remote pause flag		
	trb	sppause3	; sent an XON: clear remote pause flag
?end:	rts			; done
?srp:	tsb	sppause3	; set remote pause flag
	rts
?cklp:	bit	sppause3	; local pause is set ?
	bvs	?off		; yes, no tx possible at this time
	INDEX16
	ldx	obufcnt3	; count of bytes in output buffer
	bne	?snd		; output buffer inot empty: send next byte
?off:	lda	#00001100B	; disable TX interrupt for avoid cont. int.
	trb	!ACIACMD
	lda	#00001000B
	tsb	!ACIACMD
	bra	?done
?snd:	dex			; update count 
	stx	obufcnt3
	ldx	obufhead3	; pointer to head of out buffer
	lda	>SOBUFADDR3,x	; get data from output buffer
	sta	.ABS.ACIADR	; send data
	inx			; update head pointer
	CPU16
	txa
	and	#(SOBUFSIZ3-1)	; circular queue
	sta	obufhead3
?done:	CPU08
	;rts

; 16C550 interrupt
; ISR routine (prologue and epilogue code for context switching not showed here)
; handle 16C550 interrupt
; WARNING: assume DBR = 0 and DP saved by interrupt handler prologue code	
uartisr:
	lda	!UART_IIR
	sta	uartiir		; save for late check FIFO enabled
	lsr	a		; if bit <0> = 1 no interrupt
	bcc	$+3		; interrupt pending
	rts
	and	#0000011B	; mask on priority code
	asl	a
	tax
;	ldy	#40
;?00:	dey
;	bne	?00
	; RX timeout interrupt and RX data available interrupt use the same routine
	jmp	(?uarttbl,x)	; jump to right routine

?uarttbl:
	.DW	?msr, ?tx, ?rx, ?lsr

?lsr:	jsr	uartilsr
	;rts
	bra	uartisr		; check again interrupt (was a test)
	
?msr:	jsr	uartmsr		; MSR: modem status register
	;rts
	bra	uartisr		; check again iterrupt
	
?tx:	jsr	uarttx		; TX FIFO or THR is empty
	;rts
	bra	uartisr		; check again interrupt

?rx:	jsr	uartrx		; data available in RX FIFO or RHR
	;rts
	bra	uartisr		; check again interrupt

; this interrupt is raised when one of the modem lines change state
; here just DSR and CTS are used: DSR for signaling remote disconnession,
; CTS for hardware handshake (put local TX in pause)
uartmsr:
	lda	!UART_MSR	; clear interrupt
	tax
	asl	a
	asl	a		; <7>: /DSR, <6>: /CTS
	eor	#$C0
	sta	splin4		; update line status
	bit	spmode4
	bpl	?end		; no handshake: ignore line changes
	tay
	bpl	?cts		; /DSR is low, check /CTS
	lda	#$A0		; /DSR is high
	tsb	spstat4		; remote terminal disconnected
?cts:	txa
	lsr	a		; check /CTS transition
	bcc	?end		; no /CTS transition
	ldx	#$40		; local pause flag
	tya
	asl	a
	bpl	?clp		; transition /CTS high->low: clear local pause
	txa
	tsb	sppause4	; /CTS high -> set local pause
	lda	#00000010B	; disable TX interrupt
	trb	!UART_IER
	rts
?clp:	txa
	trb	sppause4	; /CTS low -> clear local pause
	lda	#00000010B	; enable TX interrupt
	tsb	!UART_IER
?end:	rts

; this interrupt is raised either when THR is empty
; or when TX FIFO is empty
uarttx:
	lda	#00000010B	; disable TX interrupt
	trb	!UART_IER
	bit	sppause4	; local pause is set ?
	bvs	?done		; yes, no tx possible at this time
	INDEX16
	ldy	obufcnt4	; count of bytes in output buffer
	beq	?done		; nothing to send at this time
	ldx	obuftail4	; pointer to tail of out buffer
	;lda	spcnt4		; set the bytes count
	lda	#16		; set the bytes count
	sta	sptmp4
?txl:	lda	>SOBUFADDR4,x	; get data from the tail of output buffer
	sta	!UART_SCP
	nop
	nop
	nop
	nop
	sta	!UART_SCP	
	nop
	nop
	nop
	nop
	nop
	nop
	sta	>UART_RXTX	; load data into FIFO/THR
	inx			; update head pointer
	ACC16
	txa
	and	#(SOBUFSIZ4-1)	; circular queue
	tax
	ACC08
	dey			; update count
	beq	?upd		; nothing else to send
	dec	sptmp4
	bne	?txl		; loop: load more bytes if FIFO enabled
?upd:	stx	obuftail4
	sty	obufcnt4
?done:	CPU08			; NOTE: tx interrupt will be re-enabled by put routine
	rts			; or by a negative /CTS transition 

; this interrupt is raised when receive a byte with some errors
uartilsr:
	lda	#00000101B	; disable LSR & RX interrupt
	trb	UART_IER
	lda	!UART_LSR	; get line status and clear interrupt
	sta	uartlsr		; save for further debugging
	lsr	a		; CF = 1 if available a new rx byte
	and	#$0F		; mask on rx errors
	;beq	uartrx1
	ora	#$80		; set rx error bit
	sta	spstat4		; set status register
	bcc	?done		; no new char received
	lda	!UART_RXTX	; get top fifo data or THR & discard
	bit	uartiir
	bvc	?done		; no FIFO enabled
	lda	#10000011B	; clear RX FIFO
	sta	!UART_FCR
?done:	rts			; after rx error, rx interrupt are disabled

; this interrupt is raised either when RHR is full or when RX FIFO reach the
; programmed trigger level (in this case 8)
uartrx:
	nop
	nop
	nop
	nop
	INDEX16
	ldy	ibufcnt4	; count of bytes in input buffer
	ldx	ibufhead4	; pointer to head of input buffer
?loop:	lda	!UART_RXTX	; get top fifo data or THR
	cpy	icntmax4	; check input buff. for remote pause condition
	bcc	?str		; below guard limit: store data
	xba			; B = received data
	lda	#00000010B	; hardware handshake...
	trb	!UART_MCR	; ...set RTS=1
	lda	#$80		; set remote pause flag
	tsb	sppause4	; bit 7=1 -> remote pause on
	xba			; A = received data
?chk:	cpy	#SIBUFSIZ4	; left room in input buffer?
	bcc	?str		; yes, store received byte
	lda	#$C0		; set bit 7: rx error, bit 6: rx overrun
	bra	?sst		; discard received data & exit
?str:	sta	>SIBUFADDR4,x	; now store received data
	iny			; update bytes count
	inx			; update rx head pointer
	ACC16
	txa
	and	#(SIBUFSIZ4-1)	; circular queue
	tax
	ACC08
	bit	uartiir
	bvc	?done
	lda	!UART_LSR	; get line status
	sta	uartlsr
	lsr	a		; CF = 1 if available a new rx byte
	and	#$0F		; mask on rx errors
	bcc	?done		; no data available, re-enable rx int. & exit
	beq	?loop
	ora	#$80
?sst:	sta	spstat4
?done:	stx	ibufhead4	; save rx head pointer
	sty	ibufcnt4	; save bytes count
	INDEX08
	rts

.COMMENT @
uartrx:
	;lda	#$01		; disable RX interrupt
	;trb	UART_IER
	INDEX16
	ldy	ibufcnt4	; count of bytes in input buffer
	ldx	ibufhead4	; pointer to head of input buffer
rxloop:	lda	!UART_LSR	; get line status
	sta	uartlsr
	lsr	a		; CF = 1 if available a new rx byte
	and	#$0F		; mask on rx errors
uartrx1:
	bcc	?done		; no data available, re-enable rx int. & exit
	xba			; save error code
	
	nop
	nop
	nop
	nop
	
	lda	!UART_RXTX	; get top fifo data or THR
	xba
?tst:	beq	?rx		; no rx error pending
	ora	#$80		; set rx error bit
	bra	?sst		; set status reg. & discard received data
?rx:	xba			; A = received data
	cpy	icntmax4	; check input buff. for remote pause condition
	bcc	?str		; below guard limit: store data
	xba			; B = received data
	lda	#00000010B	; hardware handshake...
	trb	!UART_MCR	; ...set RTS=1
	lda	#$80		; set remote pause flag
	tsb	sppause4	; bit 7=1 -> remote pause on
	xba			; A = received data
?chk:	cpy	#SIBUFSIZ4	; left room in input buffer?
	bcc	?str		; yes, store received byte
	lda	#$C0		; set bit 7: rx error, bit 6: rx overrun
?sst:	sta	spstat4		; set status register
	bra	?done		; discard received data & exit
?str:	sta	>SIBUFADDR4,x	; now store received data
	iny			; update bytes count
	inx			; update rx head pointer
	ACC16
	txa
	and	#(SIBUFSIZ4-1)	; circular queue
	tax
	ACC08
	bra	rxloop		; check again if rx data available
?done:	stx	ibufhead4	; save rx head pointer
	sty	ibufcnt4	; save bytes count
?done1:	INDEX08
	;lda	#$01		; re-enable rx interrupt
	;tsb	UART_IER
	rts
@

.COMMENT @
uart:
	lda	!UART_IIR
	sta	uartiir
	lsr	a		; if <0> = 1 no interrupt
	bcc	$+3		; interrupt pending
	rts
	and	#0000011B	; mask on priority code
	asl	a
	tax
	jmp	(?uarttbl,x)

?uarttbl:
	.DW	?msr, ?tx, ?rx, ?lsr


	; int. 3: line status register (LSR)
	;
?lsr:	;jsr	uartrx
	lda	!UART_LSR	; read LSR and clear interrupt
	sta	uartlsr
	;rts
	bra	uart		; check again interrupt
	
?msr:	jsr	uartmsr		; MSR: modem status register
	;rts
	bra	uart		; check again iterrupt
	
?tx:	jsr	uarttx		; TX FIFO is empty
	;rts
	bra	uart		; check again interrupt

?rx:	jsr	uartrx		; data available in RX FIFO
	;rts
	bra	uart		; check again interrupt


uartmsr:
	lda	!UART_MSR
	tax
	asl	a
	asl	a		; <7>: /DSR, <6>: /CTS
	eor	#$C0
	sta	splin4		; update line status
	tay
	bit	spmode4
	bpl	?end		; no handshake
	bvc	?end		; software handshake: no /CTS and /DSR check	
	tya
	bpl	?cts		; /DSR is low
	lda	#$A0
	tsb	spstat4		; remote terminal disconnected
?cts:	bvc	?end		; handshake software: no check /CTS
	txa
	lsr	a		; check /CTS transition
	bcc	?end		; no /CTS transition
	ldx	#$40		; local pause flag
	tya
	asl	a
	bpl	?clp		; transition high->low
	txa
	tsb	sppause4	; /CTS high set local pause
	rts
?clp:	txa
	trb	sppause4	; /CTS low clear local pause
?end:	rts

uarttx:
	;;lda	!UART_IIR

	lda	!UART_LSR
	and	#$40
	;;cmp	#$60
	bne	?cklp
	lda	!UART_IIR
	rts
	;lda	#$20
	;bit	!UART_LSR
	;bne	?cklp
	;rts
	
	bra	?cklp
	ldx	spout4		; pending an XON/XOFF sending?
	beq	?cklp		; no... check local pause flag
	
	; if send an XON then we will re-enable tx interrupt
	; else we will disable tx interrupt
	;
	lda	#00000010B	; IER<1>: tx interrupt
	cpx	#SPXON
	bne	?di
	tsb	!UART_IER	; enable tx interrupt
	bra	?htx1
?di:	trb	!UART_IER	; disable tx interrupt
?htx1:	stx	.ABS.UART_RXTX	; send XON/XOFF
	stz	spout4		; clear XON/XOFF flag
	lda	#$80		; remote pause flag
	cpx	#SPXOFF
	beq	?srp		; sent an XOFF: set remote pause flag		
	trb	sppause4	; sent an XON: clear remote pause flag
	rts			; done
?srp:	tsb	sppause4	; set remote pause flag
	rts
?cklp:	bit	sppause4	; local pause is set ?
	bvs	?off		; yes, no tx possible at this time
	INDEX16
	ldx	obufhead4	; pointer to head of out buffer
	ldy	obufcnt4	; count of bytes in output buffer
	bne	?snd		; output buffer is not empty
?off:	lda	#00000010B	; disable TX interrupt for avoid cont. int.
	trb	!UART_IER
	bra	?done
?snd:	
	lda	#00000010B	; disable TX interrupt for avoid cont. int.
	trb	!UART_IER
	
	lda	#1		; fill fifo
	sta	spout4
	lda	#$20
	bit	spmode4
	bne	?txl
	lda	#1
	sta	spout4
?txl:
	lda	>SOBUFADDR4,x	; get data from output buffer
	sta	!UART_RXTX	; send data to FIFO
	inx			; update head pointer
	ACC16
	txa
	and	#(SOBUFSIZ4-1)	; circular queue
	tax
	ACC08
	dey
	beq	?upd
	dec	spout4
	bne	?txl
?upd:	stz	spout4
	stx	obufhead4
	sty	obufcnt4
?done:	CPU08
	lda	!UART_IIR
	rts

uartrx:
	;lda	spstat4
	;bmi	?rts

	lda	#$01
	trb	UART_IER
	INDEX16
	ldy	ibufcnt4
	ldx	ibuftail4
?lp1:	lda	!UART_LSR
	sta	uartlsr
	lsr	a
	bcc	?done
	
	;bcs	?loop
	;and	#$0F
	;beq	?done1
	;ora	#$80		; set rx error bit
	;sta	spstat4
	;bra	?done1

?loop:	and	#$0F		; mask on errors	
	xba
	lda	!UART_RXTX	; get top fifo data
	xba
	beq	?rx		; no rx error pending
	ora	#$80		; set rx error bit
	bra	?sst		; set status reg. & discard received data
?rx:	xba			; A = received data
	bit	spmode4		; test for active handshake
	bpl	?chk		; no handshake: goto to check input queue
	bvs	?tst		; hardware handshake (check queue overflow)
	cmp	#SPXON		; received an XON control byte?
	bne	?xoff		; no, check if received an XOFF
	lda	#$40		; clear local pause flag
	trb	sppause4	; bit<6> = 0 -> local pause off (resume tx)
	bra	?cnt		; discard received data
?xoff:	cmp	#SPXOFF		; received an XOFF control byte?
	bne	?tst		; no, so check queue overflow
	lda	#$40		; set local pause flag
	tsb	sppause4	; bit<6> = 1 -> local pause on (will stop tx)
	bra	?cnt		; discard received data
?tst:	cpy	icntmax4	; check input buff. for remote pause condition
	bcc	?str		; below guard limit: store data
	bit	sppause4	; remote pause is already on ?
	bmi	?chk		; yes, so check input buffer
	xba			; B = received data
	bit	spmode4		; test handshake type
	bvs	?rtsh		; bit 6=1 -> hardware handshake so set RTS=1
	lda	#SPXOFF		; send an XOFF to remote terminal
	sta	spout4		; XOFF sending deffered until THR is empty
	lda	#00000010B	; IER<1>: tx interrupt
	tsb	!UART_IER	; enable tx interrupt
	bra	?xba		; check input buffer		
?rtsh:	lda	#00000010B	; hardware handshake...
	trb	!UART_MCR	; ...set RTS=1
	lda	#$80		; set remote pause
	tsb	sppause4	; bit 7=1 -> remote pause on
?xba:	xba			; A = received data
?chk:	cpy	#SIBUFSIZ4	; left room in input buffer?
	bcc	?str		; yes, store received byte
	lda	#$C0		; set bit 7: rx error, bit 6: rx overflow
?sst:	sta	spstat4		; set status register
	bra	?done		; discard received data & exit
?str:	sta	>SIBUFADDR4,x	; now store received data
	iny			; update bytes count
	inx			; update rx tail pointer
	ACC16
	txa
	and	#(SIBUFSIZ4-1)	; circular queue
	tax
	ACC08
?cnt:	;lda	!UART_LSR	; any rx data pending?
	;sta	uartlsr
	;lsr	a
	;bcs	?loop		; yes, get again from top fifo
	bra	?lp1
?done:	stx	ibuftail4	; save rx tail pointer
	sty	ibufcnt4	; save bytes count
?done1:	INDEX08
?rts:	
	lda	#$01
	tsb	UART_IER
	rts
@

int5sr:
int6sr:
	rts

; INT from RTC
int7sr:
	ldx	#RTCCTRLA	; setta banco 1 RTC
	stx	RTCALE
	lda	RTCDATA		; salva flag banco
	pha
	ora	#$10		; setta banco 1
	sta	RTCDATA
	ldx	#RTCSMI2	; recupera eventuale ALE
	stx	RTCALE
	lda	RTCDATA
	pha			; ALE da ripristinare
	ldx	#RTCSTATUS
	stx	RTCALE
	lda	RTCDATA		; azzera flag
	bmi	?01
	jmp	?20		; NO IRQ da RTC
?01:	bit	#$20		; flag AF (alarm)
	beq	?10		; no alarm
	ldx	#RTCEXTCTRLA	; check INCR bit 6
	stx	RTCALE
?02:	bit	RTCDATA
	bvs	?02
	ldx	#RTCDAY		; day of week
	stx	RTCALE
	lda	RTCDATA
	cmp	#1		; sunday?
	bne	?10		; no
	inx
	stx	RTCALE
	lda	RTCDATA		; day of month (date)
	cmp	#25		; last sunday of month?
	bcc	?10		; no
	inx
	stx	RTCALE		; month
	lda	RTCDATA
	cmp	#3		; march?
	bne	?04		; no
	ldx	#RTCHOURS	; increment hours
	stx	RTCALE
	inc	RTCDATA
	ldx	#RTCRAM0F
	stx	RTCALE
	stz	RTCDATA		; reset flag for next time	
	bra	?10
?04:	cmp	#10		; october?
	bne	?10		; no
	ldx	#RTCRAM0F
	stx	RTCALE
	lda	RTCDATA		; already updated?
	bpl	?06		; no
	stz	RTCDATA		; yes...so reset flag
	bra	?10
?06:	lda	#$FF
	sta	RTCDATA		; set updated flag
	ldx	#RTCHOURS	; decrement hours
	stx	RTCALE
	dec	RTCDATA
?10:	ldx	#RTCEXTCTRLA	; check others interrupt's
	stx	RTCALE
	lda	RTCDATA
	bit	#$01		; KF ?
	php	
	lda	#$07		; cancella altri IRQ
	trb	RTCDATA
	plp
	beq	?20		; NO KF
	lda	#$01		; TEST KS START
	bit	RTCFlag
	bne	?11		; KS OFF
	tsb	RTCFlag		; segnala KS OFF
	bra	?20
?11:	lda	#$08		; PAB = 1, PWR OFF
	;tsb	RTCDATA
	trb	RTCDATA
	lda	#$01		; CLEAR KS START
	;trb	SysRTC		
?20:	ply			; ALE da ripristinare
	ldx	#RTCCTRLA	; setta banco originale RTC
	stx	RTCALE
	pla			; banco originale
	sta	RTCDATA		; salva flag banco
	sty	RTCALE		; ripristina ALE
	rts
	

	.CODE

	.EXTERN SysStart
	.GLOBAL intsr

;
;	FWE	XFE	R/W	BANK $F0
;	0	X	X	RAM
;	1	0	X	FLASH
;	1	1	1	EEROM8U
;	1	1	0	RAM
;
; codice di startup
_RSTEntry:

	; at start: XFE=0, FWE=0, HIM=0
	sei
	cld
	;sta	!CRHIMOFF	; forza start da flash
	clc
	xce			; CPU in modo nativo
	ldy	#0
	CPU16
	ldx	#STACK_ADDR	; init stack
	txs
	tya			; DP = 0 (page 0)
	tcd
	tyx
	CPU08
	phk
	plb			; DBR = PBR = 0

	lda	!CRBIT0		; test boot
	beq	?09		; firmware from FLASH

	; check EEROM emulator board
?00:	sty	.ABS.VIA4+VIADDRB
	cpy	.ABS.VIA4+VIADDRB
	bne	?09		; fail -- goto old way
	iny
	bne	?00
	
	; ok, seem present
	dey
	sty	.ABS.VIA4+VIADDRB	; port B out

	; transfer emulator data to $F0 bank
	lda	#$F0
	sta	<2
	stz	<0
	stz	<1
	stz	!VIA4+VIAPRB
	sta	!CREMEON
	INDEX16
?lp1:	ldy	#0
	tyx
?lp2:	lda	>EMURAM,x
	sta	[0],y
	iny
	inx
	cpx	#$0800
	bcc	?lp2
	ACC16CLC
	txa
	adc	<0
	sta	<0
	ACC08
	bcc	?nxt
	inc	<2
?nxt:	inc	!VIA4+VIAPRB
	bne	?lp1
	CPU08
	bra	?10

	;sta	!CREMEOFF
	;ldy	#0
	;CPU16
	;tyx
	;lda	#$FFFF
	;mvn	#$E8, #$F8
	;mvn	#$E9, #$F9
	;mvn	#$EA, #$FA
	;mvn	#$EB, #$FB
	;mvn	#$EC, #$FC
	;mvn	#$ED, #$FD
	;mvn	#$EE, #$FE
	;mvn	#$EF, #$FF
	;CPU08
	;bra	?11

	; original boot sequence
?09:	sta	!CRFWEON	; enable FLASH/EMU IN BANK $F0
	;lda	!CRBIT0		; test boot
	beq	?10		; firmware from flash
	sta	!CRXFEON	; enable EEROM8U
?10:	ldy	#0
	CPU16
	tyx
	lda	#$FFFF
	mvn	#$F0, #$F8
	mvn	#$F1, #$F9
	mvn	#$F2, #$FA
	mvn	#$F3, #$FB
	mvn	#$F4, #$FC
	mvn	#$F5, #$FD
	mvn	#$F6, #$FE
	mvn	#$F7, #$FF
	CPU08
?11:	phk
	plb			; DBR = PBR = 0
	sta	CRXFEOFF	; disabilita flash F00000 - F7FFFF
	sta	CRFWEOFF
	jml	_SysStart	; long jmp - bootstrap (banco $F8)

_Unexpected:
	rti
	
; software interrupt generated by 'cop' istruction
;
; stack frame
;	---------
;	|  PBR  |	0F
;	---------
;	|  PCH  |	0E
;	---------	
;	|  PCL  |	0D
;	---------
;	|   P   |	0C
;	---------
;	|   B   |	0B
;	---------
;	|   A   |	0A
;	---------
;	|  XH   |	09
;	---------
;	|  XL   |	08
;	---------
;	|  YH   |	07
;	---------
;	|  YL   |	06
;	---------
;	|  DPH  |	05
;	---------
;	|  DPL  |	04
;	---------
;	|  DBR  |	03
;	---------
;	| CNTH  |	02
;	---------          --> CNT parameters bytes count 
;	| CNTL  |	01
;	---------
;
; equates for access stack offset data
STKCNT		.SET	$01
STKYR		.SET	$06
STKXR		.SET	$08
STKCR		.SET	$0A
STKBR		.SET	$0B
STKSR		.SET	$0C
STKPCL		.SET	$0D
STKPBR		.SET	$0F
STKNVAR		.SET	$0F

;USRCOP		.SET	$C0

;EINVAL		.EQU	$01

_COPEntry:
	CPU16			; A/MEM/X/Y -> 16 bit
	pha			; save C in stack
	phx			; save X(16) in stack
	phy			; save Y(16) in stack
	phd			; save DPR in stack
	phb			; save DBR in stack
	lda	#0
	tax			; X = 0
	pha			; params byte count in the stack ( = 0) 
	tcd			; set DPR = $0000
	phk			; set DBR = PBR = $00
	plb
	lda	STKPCL,s	; load PC saved in stack
	dec	a		; pointer to signature byte
	sta	COPPtr		; save long pointer
	CPU08			; A/MEM/X/Y -> 8 bit
	txa
	xba			; B = 0
	lda	STKPBR,s	; PBR bank where COP was executed
	sta	COPPtr+2
	lda	STKSR,s		; saved P in stack
	tax
	and	#~PCFLAG	; clear carry in saved P
	sta	STKSR,s
	bit	#PIFLAG		; check if IRQ was enabled
	bne	?04
	cli			; enable IRQ
?04:	phx			; saved P in stack
	lda	[COPPtr]	; fetch signature byte
	sta	COPIdx
	CPU16			; now C = $00XX
	asl	a		; index * 4	
	asl	a
	tax
	cpx	#$0200
	bcc	?05		; range $00..$7F
	lda	>SYSTBLE_ADDR,x	; address of cop vector range $80..$FF
	tay
	lda	>SYSTBLE_ADDR+2,x ; A = bank of cop vector, B = valid/params
	bra	?06
?05:	ldy	!SYSTBLE_ADDR,x	; address of cop vector range $00..$7F
	lda	!SYSTBLE_ADDR+2,x ; A = bank of cop vector, B = valid/params
?06:	sty	?08+1
	CPU08
	bmi	?30		; B<7> = 1 -> invalid function
	sta	!?08+3
	xba			; A = bytes number of params
	sta	STKCNT+1,s	; take in account the 'phx' in ?04 !!
	pla			; saved P in stack
	lsr	a		; pass carry to function
?08:	jsl	$000000
	bcc	?16		; no error
	sta	STKYR,s		; return error in Y
	lda	#0
	sta	STKYR+1,s
	lda	STKSR,s		; saved P in stack
	ora 	#PCFLAG		; set carry in saved P
	sta	STKSR,s
?16:	; epilogue code	
	CPU16
	lda	STKCNT,s	; number of params bytes in the stack
	beq	?20		; no params -- skip stack cleaning
	clc
	tsc			; C = stack pointer
	adc	#STKNVAR	; add size of stack vars
	tax			; source pointer for data move  
	adc	STKCNT,s	; add params bytes count
	tay			; dest pointer for data move
	lda	#STKNVAR-1	; move bytes count
	mvp	#0, #0		; cleanup stack
	tya			; new stack pointer
	tcs
?20:	pla			; skip STKCNT
	plb			; restore DBR
	pld			; restore DPR
	ply			; restore Y
	plx			; restore X
	pla			; restore C
	rti			; restore P and return

?30:	plx	
	jmp	?16

_BRKEntry:
	CPU16			; salva registri 16 bit
	pha			; salva A
	phx			; salva X
	phy			; salva Y
	phd
	phb			; salva DBR
	phk			; DBR = 0 = PBR(K)
	plb			; DATA BANK 0
	CPU08
	jml	BRKMON
	
; gestione interrupt non mascherabile linea NMI (RTI restore P and register size)
_NMIEntry:

	.EXTERN shutdown

	CPU16			; salva registri 16 bit
	pha			; salva A
	phx			; salva X
	phy			; salva Y
	phd
	phb			; salva DBR
	phk			; DBR = 0 = PBR(K)
	plb			; DATA BANK 0
	lda	#0		; pagina 0
	tcd
	CPU08			; imposta registri 8 bit

	lda	!VIA1+VIAIFR	; flag IFR VIA 1
	bpl	?100		; no irq da VIA 1
	lsr	a		; test CA2 (NMI KEYB)
	bcc	?50		; no CA2
	ldx	KBFR+2
	cpx	#KB_CTRLBREAK
	bne	?02
	tay
	lda	#$80
	tsb	CtrlBrk
	tya
	bra	?40
?02:	cpx	#KB_ALTSYSREQ
	bne	?06
	ldx	#CA2IFRB
	stx	VIA1+VIAIFR
	stx	KBCLRNMI	; azzera linea INT
	tay
	lda	#$40
	tsb	CtrlBrk	
	jml	ALTSYSMON
?06:
?40:	ldx	#CA2IFRB
	stx	VIA1+VIAIFR
	stx	KBCLRNMI	; azzera linea INT
?50:	lsr	a
	lsr	a
	lsr	a
	bcc	?100
	ldx	#CB2IFRB	; trigger
	stx	VIA1+VIAIFR	; clear irq flag
	lda	#$01
	bit	CtrlBrk
	bne	?100
	tsb	CtrlBrk

	
?51:	;lda	VIA1+VIAPRB
	;cmp	VIA1+VIAPRB
	;bne	?51
	;and	#$10
	;beq	?51

	; OFF command
	;ldx	#RTCEXTCTRLA
	;stx	RTCALE
	;lda	#$08		; POWERON -> HIGH	
	;tsb	RTCDATA
	
	jsl	shutdown
	lda	#$01
	trb	CtrlBrk
	

?100:	
	CPU16			; ripristina registri 16 bit
	plb			; ripristina DBR
	pld
	ply			; ripristina Y
	plx			; ripristina X
	pla			; ripristina A
	rti
	LONG_OFF

; gestione interrupt linea IRQ (RTI restore P and register size)
_IRQEntry:
	CPU16			; salva registri 16 bit
	pha			; salva A
	phx			; salva X
	phy			; salva Y
	phd			; salva DP
	phb			; salva DBR
	phk			; DBR = 0 = PBR(K)
	plb			; DATA BANK 0
	lda	#0
	tcd			; imposta pagina 0
	CPU08			; imposta registri 8 bit
	ldx	IRQVECTR	; legge vettore IRQ (x2)
	jsr 	(intsr,x)	; esegue subroutine IntSr
	CPU16			; ripristina registri 16 bit
	plb			; ripristina DBR
	pld			; ripristina DP
	ply			; ripristina Y
	plx			; ripristina X
	pla			; ripristina A
	rti
	LONG_OFF
	
;----------------------------------------------------------
; ---- VETTORI CPU
;----------------------------------------------------------
	
	.SYSVECTRS
	
	; vettori modo nativo
	.DW	0		; FFE0
	.DW	0		; FFE2
	.DW	_COPEntry	; FFE4 -> COP
	.DW	_BRKEntry	; FFE6 -> BRK
	.DW	_Unexpected	; FFE8 -> ABORT (non usato)
	.DW	_NMIEntry	; FFEA -> NMI
	.DW	_Unexpected	; FFEC -> riservato
	.DW	_IRQEntry	; FFEE -> IRQ

; nel banco $F8
_SysStart:
	jmp	!SysStart
	.DB	$00
	
	; vettori modo emulazione
	.DW	_COPEntry	; FFF4 -> COP
	.DW	_Unexpected	; FFF6 -> riservato
	.DW	_Unexpected	; FFF8 -> ABORT (non usato)
	.DW	_NMIEntry	; FFFA -> NMI
	.DW	_RSTEntry	; FFFC -> RESET
	.DW	_IRQEntry	; FFFE -> IRQ

	
	.END _RSTEntry
